'use strict'

const connect = require('it-ws/client')
const withIs = require('class-is')
const toUri = require('multiaddr-to-uri')
const { AbortError } = require('abortable-iterator')
const pDefer = require('p-defer')

const debug = require('debug')
const log = debug('libp2p:websockets')
log.error = debug('libp2p:websockets:error')
const env = require('ipfs-utils/src/env')

const createListener = require('./listener')
const toConnection = require('./socket-to-conn')
const filters = require('./filters')

/**
 * @typedef {import('multiaddr').Multiaddr} Multiaddr
 */

/**
 * @class WebSockets
 */
class WebSockets {
  /**
   * @class
   * @param {object} options
   * @param {Upgrader} options.upgrader
   * @param {(multiaddrs: Array<Multiaddr>) => Array<Multiaddr>} options.filter - override transport addresses filter
   */
  //constructor ({ upgrader, filter }) {
  constructor ({ settings, upgrader, filter }) {
    if (!upgrader) {
      throw new Error('An upgrader must be provided. See https://github.com/libp2p/interface-transport#upgrader.')
    }
    this._upgrader = upgrader
    this._filter = filter
    ////////////////////////////////////////////////////////
    if (!('WebSocket' in window)) {
      return
    }

    // Default settings
    var settings = {

      /** Whether this instance should log debug messages. */
      debug: false,

      /** The number of milliseconds to delay before attempting to reconnect. */
      reconnectInterval: 3000, //1000,
      /** The maximum number of milliseconds to delay a reconnection attempt. */
      maxReconnectInterval: 30000,
      /** The rate of increase of the reconnect delay. Allows reconnect attempts to back off when problems persist. */
      reconnectDecay: 1.5,

      /** The maximum time in milliseconds to wait for a connection to succeed before closing and retrying. */
      timeoutInterval: 5000, //2000,

      /** The maximum number of reconnection attempts to make. Unlimited if null. */
      maxReconnectAttempts: 5, //null,

      /** The binary type, possible values 'blob' or 'arraybuffer', default 'blob'. */
      binaryType: 'arraybuffer' //'blob'
    }
    if (!options) { options = {} }

    // Overwrite and define settings with options if they exist.
    for (var key in settings) {
      if (typeof options[key] !== 'undefined') {
        this[key] = options[key]
      } else {
        this[key] = settings[key]
      }
    }

    // These should be treated as read-only properties
    /**
    * The current state of the connection.
    * Can be one of: WebSocket.CONNECTING, WebSocket.OPEN, WebSocket.CLOSING, WebSocket.CLOSED
    * Read only.
    */
    this.readyState = WebSocket.CONNECTING

    this.forcedClose = false

    // Private state variables
    var eventTarget = document.createElement('div')

    // Wire up "on*" properties as event handlers
    eventTarget.addEventListener('open', function (event) { self.onopen(event) })
    eventTarget.addEventListener('close', function (event) { self.onclose(event) })
    eventTarget.addEventListener('connecting', function (event) { self.onconnecting(event) })
    eventTarget.addEventListener('message', function (event) { self.onmessage(event) })
    eventTarget.addEventListener('error', function (event) { self.onerror(event) })

    // Expose the API required by EventTarget
    this.addEventListener = eventTarget.addEventListener.bind(eventTarget)
    this.removeEventListener = eventTarget.removeEventListener.bind(eventTarget)
    this.dispatchEvent = eventTarget.dispatchEvent.bind(eventTarget)
    ////////////////////////////////////////////////////////
  }

  /**
   * @async
   * @param {Multiaddr} ma
   * @param {object} [options]
   * @param {AbortSignal} [options.signal] - Used to abort dial requests
   * @returns {Connection} An upgraded Connection
   */
  async dial (ma, options = {}) {
    log('dialing %s', ma)

    const socket = await this._connect(ma, options)
    const maConn = toConnection(socket, { remoteAddr: ma, signal: options.signal })
    log('new outbound connection %s', maConn.remoteAddr)

    const conn = await this._upgrader.upgradeOutbound(maConn)
    log('outbound connection %s upgraded', maConn.remoteAddr)
    return conn
  }

  /**
   * @private
   * @param {Multiaddr} ma
   * @param {object} [options]
   * @param {AbortSignal} [options.signal] - Used to abort dial requests
   * @returns {Promise<WebSocket>} Resolves a extended duplex iterable on top of a WebSocket
   */
  async _connect (ma, options = {}) {
    if (options.signal && options.signal.aborted) {
      throw new AbortError()
    }
    const cOpts = ma.toOptions()
    log('dialing %s:%s', cOpts.host, cOpts.port)

    const errorPromise = pDefer()
    const errfn = (err) => {
      const msg = `connection error: ${err.message}`
      log.error(msg)

      errorPromise.reject(err)
    }

    //const rawSocket = connect(toUri(ma), Object.assign({ binary: true }, options))
    ////////////////////////////////////////////////////////
    const rawSocket = connect(toUri(ma), Object.assign({ binaryType: this.binaryType }, options))
    this.ws = rawSocket.socket
    var self = this

    var timeout = setTimeout(function () {
      if (self.debug) {
        console.debug('ReconnectingWebSocket', 'connection-timeout', self.ws.url)
      }
      timedOut = true
      localWs.close()
      timedOut = false
    }, self.timeoutInterval)

    ws.onopen = function (event) {
      clearTimeout(timeout)
      if (self.debug) {
        console.debug('ReconnectingWebSocket', 'onopen', self.ws.url)
      }
      self.protocol = ws.protocol
      self.readyState = WebSocket.OPEN
      self.reconnectAttempts = 0
      var e = generateEvent('open')
      e.isReconnect = reconnectAttempt
      reconnectAttempt = false
      eventTarget.dispatchEvent(e)
    }

    ws.onclose = function (event) {
      clearTimeout(timeout)
      ws = null
      if (self.forcedClose) {
        self.readyState = WebSocket.CLOSED
        eventTarget.dispatchEvent(generateEvent('close'))
      } else {
        self.readyState = WebSocket.CONNECTING
        var e = generateEvent('connecting')
        e.code = event.code
        e.reason = event.reason
        e.wasClean = event.wasClean
        eventTarget.dispatchEvent(e)
        if (!reconnectAttempt && !timedOut) {
          if (self.debug) {
            console.debug('ReconnectingWebSocket', 'onclose', localWs.url)
          }
          eventTarget.dispatchEvent(generateEvent('close'))
        }

        var timeout = self.reconnectInterval * Math.pow(self.reconnectDecay, self.reconnectAttempts)
        setTimeout(function () {
          self.reconnectAttempts++
          self.open(true)
        }, timeout > self.maxReconnectInterval ? self.maxReconnectInterval : timeout)
      }
    }
    ws.onmessage = function (event) {
      if (self.debug) {
        console.debug('ReconnectingWebSocket', 'onmessage', localWs.url, event.data)
      }
      var e = generateEvent('message')
      e.data = event.data
      eventTarget.dispatchEvent(e)
    }
    ws.onerror = function (event) {
      if (self.debug) {
        console.debug('ReconnectingWebSocket', 'onerror', localWs.url, event)
      }
      eventTarget.dispatchEvent(generateEvent('error'))
    }
    ////////////////////////////////////////////////////////
    if (rawSocket.socket.on) {
      rawSocket.socket.on('error', errfn)
    } else {
      rawSocket.socket.onerror = errfn
    }

    if (!options.signal) {
      await Promise.race([rawSocket.connected(), errorPromise.promise])

      log('connected %s', ma)
      return rawSocket
    }

    // Allow abort via signal during connect
    let onAbort
    const abort = new Promise((resolve, reject) => {
      onAbort = () => {
        reject(new AbortError())
        // FIXME: https://github.com/libp2p/js-libp2p-websockets/issues/121
        setTimeout(() => {
          rawSocket.close()
        })
      }

      // Already aborted?
      if (options.signal.aborted) return onAbort()
      options.signal.addEventListener('abort', onAbort)
    })

    try {
      await Promise.race([abort, errorPromise.promise, rawSocket.connected()])
    } finally {
      options.signal.removeEventListener('abort', onAbort)
    }

    log('connected %s', ma)
    return rawSocket
  }

  /**
   * Creates a Websockets listener. The provided `handler` function will be called
   * anytime a new incoming Connection has been successfully upgraded via
   * `upgrader.upgradeInbound`.
   *
   * @param {object} [options]
   * @param {http.Server} [options.server] - A pre-created Node.js HTTP/S server.
   * @param {function (Connection)} handler
   * @returns {Listener} A Websockets listener
   */
  createListener (options = {}, handler) {
    if (typeof options === 'function') {
      handler = options
      options = {}
    }

    return createListener({ handler, upgrader: this._upgrader }, options)
  }

  /**
   * Takes a list of `Multiaddr`s and returns only valid Websockets addresses.
   * By default, in a browser environment only DNS+WSS multiaddr is accepted,
   * while in a Node.js environment DNS+{WS, WSS} multiaddrs are accepted.
   *
   * @param {Multiaddr[]} multiaddrs
   * @returns {Multiaddr[]} Valid Websockets multiaddrs
   */
  filter (multiaddrs) {
    multiaddrs = Array.isArray(multiaddrs) ? multiaddrs : [multiaddrs]

    if (this._filter) {
      return this._filter(multiaddrs)
    }

    // Browser
    if (env.isBrowser || env.isWebWorker) {
      return filters.dnsWss(multiaddrs)
    }

    return filters.all(multiaddrs)
  }

  /**
   * Transmits data to the server over the WebSocket connection.
   *
   * @param data a text string, ArrayBuffer or Blob to send to the server.
   */
  send (data) {
    if (this.ws) {
      if (this.debug) {
        console.debug('ReconnectingWebSocket', 'send', this.ws.url, data)
      }
      return this.ws.send(data)
    } else {
      throw 'INVALID_STATE_ERR : Pausing to reconnect websocket'
    }
  }

  /**
   * Closes the WebSocket connection or connection attempt, if any.
   * If the connection is already CLOSED, this method does nothing.
   */
  close (code, reason) {
    // Default CLOSE_NORMAL code
    if (typeof code == 'undefined') {
      code = 1000
    }
    this.forcedClose = true
    if (this.ws) {
      this.ws.close(code, reason)
    }
  }

  /**
   * Additional public API method to refresh the connection if still open (close, re-open).
   * For example, if the app suspects bad data / missed heart beats, it can try to refresh.
   */
  refresh () {
    if (this.ws) {
      this.ws.close()
    }
  }

  /**
   * This function generates an event that is compatible with standard
   * compliant browsers and IE9 - IE11
   *
   * This will prevent the error:
   * Object doesn't support this action
   *
   * http://stackoverflow.com/questions/19345392/why-arent-my-parameters-getting-passed-through-to-a-dispatched-event/19345563#19345563
   * @param s String The name that the event should use
   * @param args Object an optional object that the event will use
   */
  generateEvent (s, args) {
    var evt = document.createEvent("CustomEvent")
    evt.initCustomEvent(s, false, false, args)
    return evt
  }
}

module.exports = withIs(WebSockets, {
  className: 'WebSockets',
  symbolName: '@libp2p/js-libp2p-websockets/websockets'
})